#!/bin/bash
#
# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.
# Note: `bash` is used for $LINENO
#


# Please customize if needed
PYTHON=python
PIP=pip
CURL=curl


# Code (not meant to be edited by users)
TEMPDIR=""
HEADER_FN="header.dump"
CURL2="$CURL --silent --dump-header $HEADER_FN"
SCHEME="https"
HOST=""        # later programmatically switched to: "IP_ADDRESS"
PORT=""        # later programmatically switched to: $PORT1, $PORT2, or $PORT3
USER=""        # later programmatically switched to: "--user admin@example.com:$0$my-secret"
CA_CERT=""     # later programmatically switched to: "--cacert ca_cert_chain.pem"
ROOT=""        # later programmatically switched to: "{+restconf}" as discovered from host-meta
CLIENT_CERT="" # later programmatically switched to: "--key $KEY --cert $CERT"

YANG_DIR=`$PIP show sztpd | grep Location | awk '{print $2}'`
YANG_DIR+="/sztpd/yang"

run_unix_cmd() {
  # $1 is the line number
  # $2 is the cmd to run
  # $3 is the expected exit code
  # $4 is the name of the output file

  #printf "Running command $1..."
  output=`$2 2>&1`
  exit_code=$?
  if [[ $exit_code -ne $3 ]]; then
    printf "failed (incorrect exit status code) on line $1.\n"
    printf "  - exit code: $exit_code (expected $3)\n"
    printf "  - command: $2\n"
    if [[ -z $output ]]; then
      printf "  - output: <none>\n\n"
    else
      printf "  - output: <starts on next line>\n$output\n\n"
    fi
    rm -f ca_cert_chain.pem
    exit
  fi

  if [[ -n "$4" ]]; then
    echo "$output" > $4
  fi
}


run_curl_cmd() {
  # $1 is the line number
  # $2 is the cmd to run
  # $3 is the expected curl exit code
  # $4 is the expected HTTP status code
  # $5 is the name of the output file

  output=`$2`
  exit_code=$?
  if [[ $exit_code -ne $3 ]]; then
    printf "failed (incorrect curl exit status code) on line $1.\n"
    printf "  - exit code: $exit_code (expected $3)\n"
    printf "  - command: $2\n"
    if [[ $exit_code -eq 7 ]]; then
      output="Failed to connect to host."
      printf "  - output: $output\n\n"
    else
      printf "  - output: <starts on next line>\n$output\n\n"
    fi
    rm -f ca_cert_chain.pem
    exit
  fi
  status_line=`grep "HTTP/" $HEADER_FN | tail -n 1`  # `tail` in case there is a "100" that needs to skipped over
  rm $HEADER_FN
  status_code=`echo "$status_line" | sed 's#HTTP/[0-9]\.[0-9] \([0-9]\{3\}\).*$#\1#'`
  if [[ $status_code -ne $4 ]]; then
    printf "failed (incorrect HTTP status code) on line $1.\n"
    printf "  - HTTP code: $status_code (expected $4)\n"
    printf "  - command: $2\n"
    printf "  - output: <starts on next line>\n$output\n\n"
    rm -f ca_cert_chain.pem
    exit
  fi
  echo "$output" > $5
}


get_path() {
  # $1 : the line number
  # $2 : the path
  # $3 : expected `curl` exit status code
  # $4 : expected HTTP status code
  # $5 : name of the output file

  #printf "GET-ing $2..."
#  echo "000"
#  command="openssl s_client -connect $HOST:$PORT --CAfile $SERVER_CA_CERT -key $CLIENT_KEY -keyform DER -cert $CLIENT_KEY_CERT -msg"
#  echo "111"
#  run_unix_cmd $1 "$command" 0 $TEMPDIR/s_client.txt
#  echo "START s_client.txt"
#  cat $TEMPDIR/s_client.txt
#  echo "STOP s_client.txt"
#  echo

  #command="$CURL2 -v $CA_CERT $CLIENT_CERT $USER $SCHEME://$HOST:$PORT$2"
  command="$CURL2 -H "Accept:application/yang-data+json" $CA_CERT $CLIENT_CERT $USER $SCHEME://$HOST:$PORT$2"
  run_curl_cmd $1 "$command" $3 $4 $5
  #printf "okay.\n"
}


post_path() {
  # $1 : the line number
  # $2 : the path
  # $3 : expected `curl` exit status code
  # $4 : expected HTTP status code
  # $5 : the name of the input file
  # $6 : the name of the output file

  #printf "POST-ing $2..."

  echo $5 | grep '\.json$' >> /dev/null 2>&1
  if [[ $? -eq 0 ]]; then
      CONTENT_TYPE="-H Content-Type:application/yang-data+json"
      ACCEPT_TYPE="-H Accept:application/yang-data+json"
  else
    echo $5 | grep '\.xml$' >> /dev/null 2>&1
    if [[ $? -eq 0 ]]; then
        CONTENT_TYPE="-H Content-Type:application/yang-data+xml"
        ACCEPT_TYPE="-H Accept:application/yang-data+xml"
    else
      echo "Unrecognized content-type suffix for: $5"
      exit
    fi
  fi

  command="$CURL2 --data @$5 $CONTENT_TYPE $ACCEPT_TYPE $CA_CERT $CLIENT_CERT $USER $SCHEME://$HOST:$PORT$2"
  run_curl_cmd $1 "$command" $3 $4 $6

  #printf "okay.\n"
}


validate_config() {
  # $1 : the line number
  # $2 : name of xml file containing the config snippet
  # $3 : name of yang file

  #printf "VALIDATE-ing config..."
  command="yanglint --verbose --strict --type config $Fs_v1 $YANG_DIR/ietf-x509-cert-to-name@2014-12-10.yang $3 $2"
  run_unix_cmd $1 "$command" 0 $TEMPDIR/validate_out
  #printf "okay.\n"
}



extract_path_value() {
  # Parameters:
  #   $1 : file containing json
  #   $2 : the path value to extract
  #
  # Returns the value, empty string otherwise

  $PYTHON - << EOF
import json
f = open("$1")
j=json.load(f)
f.close()
segments="$2".split('/')
ptr=j
for segment in segments:
  if segment not in ptr:
    print("")
  else:
    ptr=ptr[segment]
    if type(ptr) == list:
      ptr=ptr[0]
print(str(ptr))
EOF
}


extract_certs_to_pem() {
  # Parameters:
  #   $1 : b64 of trust anchor cms
  #   $2 : name of the PEM file to create

  $PYTHON - << EOF
import base64
import textwrap
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5280
from pyasn1.type import univ
b64="$1"
cms=base64.b64decode(b64)
content, rest_of_input = der_decoder.decode(cms, asn1Spec=rfc5652.ContentInfo())
if content['contentType'] != rfc5652.id_signedData:
  raise KeyError("unexpected content type: " + str(content['contentType']))
signed_data, rest_of_input = der_decoder.decode(content['content'], asn1Spec=rfc5652.SignedData())
certificates = signed_data["certificates"]
all_pems = ""
for cert in certificates:
    inner_cert = cert["certificate"]
    assert(type(inner_cert) == rfc5280.Certificate)
    der_serialisation = der_encoder.encode(inner_cert)
    pem_serialisation = "-----BEGIN CERTIFICATE-----\n"
    b64_long_string = base64.b64encode(der_serialisation).decode('ASCII')
    b64_wrapped = textwrap.fill(b64_long_string, 64)
    pem_serialisation += b64_wrapped + '\n'
    pem_serialisation += "-----END CERTIFICATE-----\n"
    all_pems += pem_serialisation
with open("$2", 'w') as pem_file:
    pem_file.write(all_pems)
EOF
}


process_redirect_information() {
  # $1 : the file constaining the redirect/conveyed information
  #
  # Return 0 (success) or 1 (fail)

  echo "  ^-- Processing redirect information..."

  # extract values from redirect-information document
  address=$(extract_path_value $1 ietf-sztp-conveyed-info:redirect-information/bootstrap-server/address)
  port=$(extract_path_value $1 ietf-sztp-conveyed-info:redirect-information/bootstrap-server/port)
  trust_anchor=$(extract_path_value $1 ietf-sztp-conveyed-info:redirect-information/bootstrap-server/trust-anchor)

  # convert CMS to a PEM containing a list of certs
  extract_certs_to_pem "$trust_anchor" $TEMPDIR/ca_certs.pem

  # override a few global variables
  HOST="$address"
  PORT="$port"
  CA_CERT="--cacert $TEMPDIR/ca_certs.pem"

  # test well-known/host-meta
  get_path $LINENO "/.well-known/host-meta" 0 200 $TEMPDIR/output.xml
  export XML_CATALOG_FILES=xrd/xmlcatalog.xml
  command="xmllint --noout --schema xrd/xrd-1.0-os.xsd $TEMPDIR/output.xml"
  run_unix_cmd $LINENO "$command" 0

  # extract restconf server root from host-meta
  ROOT=`grep 'rel="restconf"' $TEMPDIR/output.xml | sed -e 's/.*href="\(.*\)".*/\1/'`
  rm $TEMPDIR/output.xml

  # prepare 'input' for the YANG-defined RPC.
  cat << EOM > get-bootstrapping-information-trusted.json
{
  "ietf-sztp-bootstrap-server:input" : {
    "hw-model" : "model-x",
    "os-name" : "vendor-os",
    "os-version" : "17.3R2.1",
    "nonce" : "BASE64VALUE="
  }
}
EOM
  echo "  ^-- Getting bootstrapping data from next server..."
  post_path $LINENO "$ROOT/operations/ietf-sztp-bootstrap-server:get-bootstrapping-data" 0 200 get-bootstrapping-information-trusted.json $TEMPDIR/output.json
  rm get-bootstrapping-information-trusted.json

  # recurse!
  #echo "recursing with this input:"
  #cat $TEMPDIR/output.json
  process_bootstrapping_data $TEMPDIR/output.json
  return $?
}




process_onboarding_information() {
  # $1 : the file constaining the onboarding/conveyed information
  #
  # Return 0 (success) or 1 (fail)

  echo "  ^-- Processing onboarding information..."

  os_name=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/boot-image/os-name)
  os_version=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/boot-image/os-version)
  download_uri=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/boot-image/download-uri)
  hash_alg=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/boot-image/image-verification/hash-algorithm)
  hash_value=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/boot-image/image-verification/hash-value)
  pre_script=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/pre-configuration-script)
  config=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/configuration)
  config_handling=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/configuration-handling)
  post_script=$(extract_path_value $1 ietf-sztp-conveyed-info:onboarding-information/post-configuration-script)

#  echo "os_name = $os_name"
#  echo "os_version = $os_version"
#  echo "download_uri = $download_uri"
#  echo "hash_alg = $hash_alg"
#  echo "hash_value = $hash_value"
#  echo "pre_script = $pre_script"
#  echo "config = $config"
#  echo "config_handling = $config_handling"
#  echo "post_script = $post_script"

  #echo "installing boot image... (not implemented ;)"

  # base64-decode fields
  PRE_SCRIPT=$(echo $pre_script | openssl enc -d -base64 -A)
  CONFIG=$(echo $config | openssl enc -d -base64 -A)
  POST_SCRIPT=$(echo $post_script | openssl enc -d -base64 -A)
  #echo "-----START PRE-SCRIPT-----"
  #echo "$PRE_SCRIPT"
  #echo "-----END PRE-SCRIPT-----"
  #echo "-----START CONFIG-----"
  #echo "$CONFIG"
  #echo "-----END CONFIG-----"
  #echo "-----START POST-SCRIPT-----"
  #echo "$POST_SCRIPT"
  #echo "-----END POST-SCRIPT-----"

  echo "  ^-- Bootstrap complete."
  return 0
}




process_bootstrapping_data() {
  # This is a recursive function (i.e., process_redirect_information() call this function)
  #
  # Parameters:
  #   $1 : the file constaining the bootstrapping data
  #
  # Return 0 (success) or 1 (fail)

  echo "  ^-- Processing bootstrapping data..."

  # This code skips validating in incoming file because `yanglint` doesn't support RESTCONF

  # extract and validate conveyed-info
  $PYTHON - << EOF
import json
import base64
from pyasn1.codec.der import decoder
from pyasn1_modules import rfc5652
from pyasn1.type import univ
f = open("$1")
j=json.load(f)
f.close()
b64=j['ietf-sztp-bootstrap-server:output']['conveyed-information']
cms=base64.b64decode(b64)
content, rest_of_input = decoder.decode(cms, asn1Spec=rfc5652.ContentInfo())
id_ct_sztpConveyedInfoJSON = rfc5652._buildOid(1, 2, 840, 113549, 1, 9, 16, 1, 43)
if content['contentType'] != id_ct_sztpConveyedInfoJSON:
  raise KeyError("unexpected content type: " + str(content['contentType']))
payload, rest_of_input = decoder.decode(content['content'], asn1Spec=univ.OctetString())
f = open("$TEMPDIR/conveyed_info.json", "w")
f.write(str(payload))
f.close()
EOF
  validate_config $LINENO "$TEMPDIR/conveyed_info.json" "$YANG_DIR/ietf-sztp-conveyed-info@2019-04-30.yang"
  #echo "##### START CONVEYED INFO #####"
  #cat conveyed_info.json
  #echo "##### STOP CONVEYED INFO #####"

  cat $TEMPDIR/conveyed_info.json | grep -q "ietf-sztp-conveyed-info:redirect-information"
  if [[ $? -eq 0 ]]; then
    process_redirect_information $TEMPDIR/conveyed_info.json
  else
    process_onboarding_information $TEMPDIR/conveyed_info.json
  fi
  exit_status=$?
  #rm conveyed_info.json
  return $exit_status
}




send_bootstrap_initiated() {
  # JSON
  cat << EOM > report-progress-bootstrap-initiated.json
{
  "ietf-sztp-bootstrap-server:input" : {
    "progress-type" : "bootstrap-initiated",
    "message" : "message sent via JSON"
  }
}
EOM
  post_path $LINENO "$ROOT/operations/ietf-sztp-bootstrap-server:report-progress" 0 204 report-progress-bootstrap-initiated.json $TEMPDIR/output.json
  rm report-progress-bootstrap-initiated.json
}




fetch_boot_image() {
#  # FETCH THE BOOT IMAGE
#  command="curl -s --dump-header header.dump --user $USER:$PASSWORD $SCHEME://$HOST:$PORT/images/foobar-19.1R2.3.img"
#  output=`run_cmd "$command" 0 200`
#  exit_code=$?
#  if [[ $exit_code -ne 0 ]]; then
#    echo "OUTPUT = $output"
#    exit 1
#  fi
  echo "installing boot image... (not implemented yet)"
}



main() {
  # $1 : the device's serial number
  # $2 : the device's password
  # $3 : the device's private key
  # $4 : the device's end-entity cert
  # $5 : the trusted bootstrap server's address
  # $6 : the trusted bootstrap server's port
  # $7 : the trusted bootstrap server's trust-anchor certificate (PEM)

  echo "  ^-- Getting bootstrapping data..."

  # set global variables passed into `curl` calls
  USER="--user $1:$2"
  CLIENT_CERT="--key $3 --cert $4"
  SERVER_CA_CERT="$7"
  HOST="$5"
  PORT="$6"
  CA_CERT="--cacert $7"

  # set a global variable for files
  TEMPDIR=`mktemp -d`

  # test well-known/host-meta
  get_path $LINENO "/.well-known/host-meta" 0 200 $TEMPDIR/output.xml
  export XML_CATALOG_FILES=xrd/xmlcatalog.xml
  command="xmllint --noout --schema xrd/xrd-1.0-os.xsd $TEMPDIR/output.xml"
  run_unix_cmd $LINENO "$command" 0

  # extract restconf server root from host-meta
  ROOT=`grep 'rel="restconf"' $TEMPDIR/output.xml | sed -e 's/.*href="\(.*\)".*/\1/'`
  rm $TEMPDIR/output.xml

  # prepare 'input' for the YANG-defined RPC
  cat << EOM > get-bootstrapping-information-trusted.json
{
  "ietf-sztp-bootstrap-server:input" : {
    "hw-model" : "model-x",
    "os-name" : "vendor-os",
    "os-version" : "17.3R2.1",
    "nonce" : "BASE64VALUE="
  }
}
EOM

  # reach out to the globally configured bootstrap server (exits on error)
  post_path $LINENO "$ROOT/operations/ietf-sztp-bootstrap-server:get-bootstrapping-data" 0 200 get-bootstrapping-information-trusted.json $TEMPDIR/output.json
  rm get-bootstrapping-information-trusted.json

  # falling into recursive function 
  process_bootstrapping_data $TEMPDIR/output.json
  result=$?

  #echo "all done."
  exit $result
}


main "$@"

