#!/bin/bash
#
# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.
# Note: `bash` is used for $LINENO
#


# Please customize if needed
PYTHON=python
OPENSSL=openssl
DEF_PORT1=1080
DEF_PORT2=2080
DEF_PORT3=3080
NBI_PORT1=10443
NBI_PORT2=20443
NBI_PORT3=30443
SBI_PORT1=18443
SBI_PORT2=28443
SBI_PORT3=38443


# Code (not meant to be edited by users)
TEMPDIR=""
PID1=""
PID2=""
PID3=""
HEADER_FN=""
CURL=""
HOST=""
SCHEME=""
PORT=""
USER=""
CA_CERT=""
CLIENT_CERT=""
ROOT=""

YANG_DIR=`$PYTHON -c 'import pkg_resources; print(pkg_resources.resource_filename("sztpd", "yang/"))'`
DATE=`echo $YANG_DIR/wn-sztpd-1* | sed 's/[^\@]*\@\(.*\).yang/\1/'`

Fs_Common="--features=ietf-truststore:certificates --features=ietf-keystore: --feature=ietf-tcp-common: --feature=ietf-tcp-client: --feature=ietf-tcp-server: --features=ietf-tls-server:x509-certificate-auth,client-auth-config-supported --features=ietf-http-client:basic-auth --features=ietf-http-server: --features=ietf-restconf-server:http-listen,https-listen"
Fs_v1="$Fs_Common --features=wn-sztpd-0:onboarding-supported,artifact-callbacks-implemented"
Fs_vZ="$Fs_Common"
Fs_v1_bss1_impl="--features=wn-sztpd-1:"
Fs_v1_bssx_impl="--features=wn-sztpd-1:system-level-administration-disabled"
Fs_vZ_bss1_impl="--features=wn-sztpd-1:device-ownership-verification"
Fs_vZ_bssx_impl="--features=wn-sztpd-1:system-level-administration-disabled --features=wn-sztpd-x:device-ownership-verification"



run_unix_cmd() {
  # $1 is the line number
  # $2 is the cmd to run
  # $3 is the expected exit code
  # $4 is the name of the output file

  #printf "Running command $1..."
  output=`$2 2>&1`
  exit_code=$?
  if [[ $exit_code -ne $3 ]]; then
    printf "failed (incorrect exit status code) on line $1.\n"
    printf "  - exit code: $exit_code (expected $3)\n"
    printf "  - command: $2\n"
    if [[ -z $output ]]; then
      printf "  - output: <none>\n\n"
    else
      printf "  - output: <starts on next line>\n$output\n\n"
    fi
    exit 1
  fi

  if [[ -n "$4" ]]; then
    echo "$output" > $4
  fi
}


run_curl_cmd() {
  # $1 is the line number
  # $2 is the cmd to run
  # $3 is the expected curl exit code
  # $4 is the expected HTTP status code
  # $5 is the name of the output file

  output=`$2`
  exit_code=$?
  if [[ $exit_code -ne $3 ]]; then
    printf "failed (incorrect curl exit status code) on line $1.\n"
    printf "  - exit code: $exit_code (expected $3)\n"
    printf "  - command: $2\n"
    if [[ $exit_code -eq 7 ]]; then
      output="Failed to connect to host."
      printf "  - output: $output\n\n"
    else
      printf "  - output: <starts on next line>\n$output\n\n"
    fi
    exit 1
  fi
  status_line=`grep "HTTP/" $HEADER_FN | tail -n 1`  # `tail` in case there is a "100" that needs to skipped over
  rm $HEADER_FN
  status_code=`echo "$status_line" | sed 's#HTTP/[0-9]\.[0-9] \([0-9]\{3\}\).*$#\1#'`
  if [[ $status_code -ne $4 ]]; then
    printf "failed (incorrect HTTP status code) on line $1.\n"
    printf "  - HTTP code: $status_code (expected $4)\n"
    printf "  - command: $2\n"
    printf "  - output: <starts on next line>\n$output\n\n"
    exit 1
  fi
  echo "$output" > $5
}


get_path() {
  # $1 : the line number
  # $2 : the path
  # $3 : expected `curl` exit status code
  # $4 : expected HTTP status code
  # $5 : name of the output file

  #printf "GET-ing $2..."
  command="$CURL -H "Accept:application/yang-data+json" $CA_CERT $USER $SCHEME://$HOST:$PORT$2"
  run_curl_cmd $1 "$command" $3 $4 $5
  #printf "okay.\n"
}



put_path() {
  # $1 : the line number
  # $2 : the path
  # $3 : expected `curl` exit status code
  # $4 : expected HTTP status code
  # $5 : the name of the input file
  # $6 : the name of the output file

  #printf "PUT-ing $2..."
  echo $5 | grep '\.json$' >> /dev/null 2>&1
  if [[ $? -eq 0 ]]; then
      CONTENT_TYPE="-H Content-Type:application/yang-data+json"
      ACCEPT_TYPE="-H Accept:application/yang-data+json"
  else
    echo $5 | grep '\.xml$' >> /dev/null 2>&1
    if [[ $? -eq 0 ]]; then
        CONTENT_TYPE="-H Content-Type:application/yang-data+xml"
        ACCEPT_TYPE="-H Accept:application/yang-data+xml"
    else
      echo "Unrecognized content-type suffix for: $5"
      exit
    fi
  fi
  command="$CURL -X PUT --data @$5 $CONTENT_TYPE $ACCEPT_TYPE $CA_CERT $USER $SCHEME://$HOST:$PORT$2"
  run_curl_cmd $1 "$command" $3 $4 $6
  #printf "okay.\n"
}



validate_config() {
  # $1 : the line number
  # $2 : name of xml file containing the config snippet
  # $3 : name of yang file
  #printf "VALIDATE-ing config..."
  command="yanglint --verbose --strict --type config $Fs_v1 $Fs_v1_bss1_impl $YANG_DIR/ietf-crypto-types@$DATE.yang $YANG_DIR/ietf-x509-cert-to-name@2014-12-10.yang $YANG_DIR/wn-x509-c2n@$DATE.yang $YANG_DIR/ietf-sztp-conveyed-info@2019-04-30.yang $3 $2"
  run_unix_cmd $1 "$command" 0 $TEMPDIR/validate_out
  #printf "okay.\n"
}



create_sztpd_instance() {
  # $1 : the instance number
  printf "   ^-- Creating SZTPD instance $1..."
  DEFAULT_PORT="808$1"
  case "$1" in
    "1")
      SZTPD_INIT_PORT="$DEF_PORT1"
      ;;
    "2")
      SZTPD_INIT_PORT="$DEF_PORT2"
      ;;
    "3")
      SZTPD_INIT_PORT="$DEF_PORT3"
      ;;
  esac 
  export SZTPD_INIT_PORT="$SZTPD_INIT_PORT"
  export SZTPD_INIT_MODE=1;
  export SZTPD_ACCEPT_CONTRACT="Yes";
  #sztpd sqlite:///:memory: & disown
  sztpd sqlite:///:memory: > $TEMPDIR/sztpd_pid$1 2>&1 & disown
  local PID="$!"
  export -n SZTPD_INIT_PORT=
  case "$1" in
    "1")
      PID1="$PID"
      ;;
    "2")
      PID2="$PID"
      ;;
    "3")
      PID3="$PID"
      ;;
  esac 
  printf "okay.  (SZTPD instance $1 running with PID $PID)\n"
}




configure_sztpd_instance() {
  # $1 : the instance number
  echo "   ^-- Configuring SZTPD instance $1..."

  # construct config to push to SZTPD server

  # set the common template variables
  case "$1" in
    "1")
      NBI_PORT="$NBI_PORT1"
      SBI_PORT="$SBI_PORT1"
      ;;
    "2")
      NBI_PORT="$NBI_PORT2"
      SBI_PORT="$SBI_PORT2"
      ;;
    "3")
      NBI_PORT="$NBI_PORT3"
      SBI_PORT="$SBI_PORT3"
      ;;
  esac 
  NBI_PRI_KEY_B64=`$OPENSSL enc -base64 -A -in pki/sztpd$1/nbi/end-entity/private_key.der`
  NBI_PUB_KEY_B64=`$OPENSSL enc -base64 -A -in pki/sztpd$1/nbi/end-entity/public_key.der`
  cat pki/sztpd$1/nbi/end-entity/my_cert.pem pki/sztpd$1/nbi/intermediate2/my_cert.pem > $TEMPDIR/cert_chain.pem
  $OPENSSL crl2pkcs7 -nocrl -certfile $TEMPDIR/cert_chain.pem -outform DER -out $TEMPDIR/cert_chain.cms
  NBI_EE_CERT_B64=`$OPENSSL enc -base64 -A -in $TEMPDIR/cert_chain.cms`
  SBI_PRI_KEY_B64=`$OPENSSL enc -base64 -A -in pki/sztpd$1/sbi/end-entity/private_key.der`
  SBI_PUB_KEY_B64=`$OPENSSL enc -base64 -A -in pki/sztpd$1/sbi/end-entity/public_key.der`
  cat pki/sztpd$1/sbi/end-entity/my_cert.pem pki/sztpd$1/sbi/intermediate2/my_cert.pem > $TEMPDIR/cert_chain.pem
  $OPENSSL crl2pkcs7 -nocrl -certfile $TEMPDIR/cert_chain.pem -outform DER -out $TEMPDIR/cert_chain.cms
  SBI_EE_CERT_B64=`$OPENSSL enc -base64 -A -in $TEMPDIR/cert_chain.cms`
  cat pki/client/root-ca/my_cert.pem pki/client/intermediate1/my_cert.pem pki/client/intermediate2/my_cert.pem > $TEMPDIR/ta_cert_chain.pem
  $OPENSSL crl2pkcs7 -nocrl -certfile $TEMPDIR/ta_cert_chain.pem -outform DER -out $TEMPDIR/ta_cert_chain.cms
  CLIENT_CERT_TA_B64=`$OPENSSL enc -base64 -A -in $TEMPDIR/ta_cert_chain.cms`

  sed -e "s#\$NBI_PORT#$NBI_PORT#" -e "s#\$SBI_PORT#$SBI_PORT#" -e "s#\$NBI_PRI_KEY_B64#$NBI_PRI_KEY_B64#" \
    -e "s#\$NBI_PUB_KEY_B64#$NBI_PUB_KEY_B64#" -e "s#\$NBI_EE_CERT_B64#$NBI_EE_CERT_B64#" \
    -e "s#\$SBI_PRI_KEY_B64#$SBI_PRI_KEY_B64#" -e "s#\$SBI_PUB_KEY_B64#$SBI_PUB_KEY_B64#" \
    -e "s#\$SBI_EE_CERT_B64#$SBI_EE_CERT_B64#" -e "s#\$CLIENT_CERT_TA_B64#$CLIENT_CERT_TA_B64#" \
    templates/common-server-configuration-template.json > $TEMPDIR/server-configuration.json

  # set the redirect or onboarding specific variables
  case "$1" in
    "1" | "2")
      NEXT=`expr $1 + 1`
      NEXT_PORT="SBI_PORT$NEXT"
      BOOTSVR_PORT=${!NEXT_PORT}
      cat pki/sztpd$NEXT/sbi/root-ca/my_cert.pem pki/sztpd$NEXT/sbi/intermediate1/my_cert.pem > $TEMPDIR/cert_chain.pem
      $OPENSSL crl2pkcs7 -nocrl -certfile $TEMPDIR/cert_chain.pem -outform DER -out $TEMPDIR/cert_chain.cms # >> /dev/null 2>&1
      BOOTSVR_TA_CERT_B64=`$OPENSSL enc -base64 -A -in  $TEMPDIR/cert_chain.cms`
      sed -e "s#\$BOOTSVR_PORT#$BOOTSVR_PORT#" -e "s#\$BOOTSVR_TA_CERT_B64#$BOOTSVR_TA_CERT_B64#" \
        templates/redirect-server-configuration-template.json >> $TEMPDIR/server-configuration.json
      ;;
    "3")
      # create a random file
      #echo "   ^-- FIXME: post bootimage file to onboarding server...?"
      dd if=/dev/urandom of=$TEMPDIR/my-boot-image.img bs=64x1024x1024 count=1 >> /dev/null 2>&1
      cat << EOM > $TEMPDIR/my-pre-configuration-script
#/bin/bash
echo "inside the pre-configuration-script..."
EOM
      cat << EOM > $TEMPDIR/my-post-configuration-script
#/bin/bash
echo "inside the post-configuration-script..."
EOM
      cat << EOM > $TEMPDIR/my-configuration
<top xmlns="https:/example.com/config">
  <any-xml-content-okay/>
</top>
EOM
      BOOT_IMG_HASH_ALG="ietf-sztp-conveyed-info:sha-256"
      BOOT_IMG_HASH_VAL=`$OPENSSL dgst -sha256 -c $TEMPDIR/my-boot-image.img | awk '{print $2}'`
      PRE_SCRIPT_B64=`$OPENSSL enc -base64 -A -in $TEMPDIR/my-pre-configuration-script`
      POST_SCRIPT_B64=`$OPENSSL enc -base64 -A -in $TEMPDIR/my-post-configuration-script`
      CONFIG_B64=`$OPENSSL enc -base64 -A -in $TEMPDIR//my-configuration`
      sed -e "s#\$BOOT_IMG_HASH_ALG#$BOOT_IMG_HASH_ALG#" -e "s#\$BOOT_IMG_HASH_VAL#$BOOT_IMG_HASH_VAL#" \
        -e "s#\$PRE_SCRIPT_B64#$PRE_SCRIPT_B64#" -e "s#\$POST_SCRIPT_B64#$POST_SCRIPT_B64#" \
        -e "s#\$CONFIG_B64#$CONFIG_B64#" templates/onboarding-server-configuration-template.json \
        >> $TEMPDIR/server-configuration.json
      ;;
  esac 
  validate_config $LINENO "$TEMPDIR/server-configuration.json" "$YANG_DIR/wn-sztpd-1@$DATE.yang"


  # init global vars
  SCHEME="http"
  HOST="127.0.0.1"
  case "$1" in
    "1")
      PORT="$DEF_PORT1"
      ;;
    "2")
      PORT="$DEF_PORT2"
      ;;
    "3")
      PORT="$DEF_PORT3"
      ;;
  esac 
  USER=""
  CA_CERT=""
  CLIENT_CERT=""
  ROOT=""

  # set the ROOT global variable
  get_path $LINENO "/.well-known/host-meta" 0 200 $TEMPDIR/output.xml
  export XML_CATALOG_FILES=xrd/xmlcatalog.xml
  command="xmllint --noout --schema xrd/xrd-1.0-os.xsd $TEMPDIR/output.xml"
  run_unix_cmd $LINENO "$command" 0
  ROOT=`grep 'rel="restconf"' $TEMPDIR/output.xml | sed -e 's/.*href="\(.*\)".*/\1/'`

  # set the sztpd server's configuration (this will cause the sztpd server to restart)
  put_path $LINENO "$ROOT/ds/ietf-datastores:running" 0 204 "$TEMPDIR/server-configuration.json" $TEMPDIR/output.json
}



test_sztpd_instance() {
  # $1 : the instance number

  echo "   ^-- Testing SZTPD instance $1"

  # init global vars
  SCHEME="https"
  HOST="127.0.0.1"
  case "$1" in
    "1")
      PORT="$NBI_PORT1"
      ;;
    "2")
      PORT="$NBI_PORT2"
      ;;
    "3")
      PORT="$NBI_PORT3"
      ;;
  esac 
  USER="--user my-admin@example.com:my-secret"
  cat pki/sztpd$1/nbi/root-ca/my_cert.pem pki/sztpd$1/nbi/intermediate1/my_cert.pem > $TEMPDIR/ca_cert_chain.pem
  CA_CERT="--cacert $TEMPDIR/ca_cert_chain.pem"
  CLIENT_CERT=""
  ROOT=""

  # set the ROOT global variable
  get_path $LINENO "/.well-known/host-meta" 0 200 $TEMPDIR/output.xml
  ROOT=`grep 'rel="restconf"' $TEMPDIR/output.xml | sed -e 's/.*href="\(.*\)".*/\1/'`

  # test config
  get_path $LINENO "$ROOT/ds/ietf-datastores:running" 0 200 $TEMPDIR/output.json
  validate_config $LINENO "$TEMPDIR/output.json" "$YANG_DIR/wn-sztpd-1@$DATE.yang"
}



test_sztpd_instance_after() {
  # $1 : the instance number

  echo "   ^-- Testing SZTPD instance $1"

  # init global vars
  SCHEME="https"
  HOST="127.0.0.1"
  case "$1" in
    "1")
      PORT="$NBI_PORT1"
      ;;
    "2")
      PORT="$NBI_PORT2"
      ;;
    "3")
      PORT="$NBI_PORT3"
      ;;
  esac 
  USER="--user my-admin@example.com:my-secret"
  cat pki/sztpd$1/nbi/root-ca/my_cert.pem pki/sztpd$1/nbi/intermediate1/my_cert.pem > $TEMPDIR/ca_cert_chain.pem
  CA_CERT="--cacert $TEMPDIR/ca_cert_chain.pem"
  CLIENT_CERT=""
  ROOT=""

  # set the ROOT global variable
  get_path $LINENO "/.well-known/host-meta" 0 200 $TEMPDIR/output.xml
  ROOT=`grep 'rel="restconf"' $TEMPDIR/output.xml | sed -e 's/.*href="\(.*\)".*/\1/'`

  # test config
  get_path $LINENO "$ROOT/ds/ietf-datastores:operational/wn-sztpd-1:devices/device=my-serial-number/bootstrapping-log" 0 200 $TEMPDIR/output.json
  echo "bootstrapping log: <next line>"
  cat "$TEMPDIR/output.json"
  echo ""
  #validate_config $LINENO "$TEMPDIR/output.json" "$YANG_DIR/wn-sztpd-1@$DATE.yang"
}




kill_sztpd_instance() {
  # $1 : the instance number
  case "$1" in
    "1")
      PID="$PID1"
      ;;
    "2")
      PID="$PID2"
      ;;
    "3")
      PID="$PID3"
      ;;
  esac 
  echo "   ^-- Sending SIGTERM to instance $1 (PID $PID)"
  kill -s TERM $PID
}



verify_instance_killed() {
  # $1 : the instance number
  printf "   ^-- Verifying SZTPD instance $1 killed..."
  case "$1" in
    "1")
      PID="$PID1"
      ;;
    "2")
      PID="$PID2"
      ;;
    "3")
      PID="$PID3"
      ;;
  esac 
  if ps -p $PID > /dev/null 2>&1
  then
     printf 'error! ($PID is still running! (please `kill` is manually)\n'
  else
     printf "okay.\n"
  fi
}



run_rfc8572_agent() {
  # this function takes no parameters

  CLIENT_SERIAL_NUMBER="my-serial-number"
  CLIENT_PASSWORD="my-secret"
  CLIENT_CERT_KEY="pki/client/end-entity/private_key.pem" # was der on Jan 23
  CLIENT_CERT_EE="pki/client/end-entity/my_cert.pem"
  BOOTSVR_ADDRESS="127.0.0.1"
  BOOTSVR_PORT="$SBI_PORT1"
  BOOTSVR_TA_CERT=$TEMPDIR/trust_chain.pem
  cat pki/sztpd1/sbi/root-ca/my_cert.pem pki/sztpd1/sbi/intermediate1/my_cert.pem > $TEMPDIR/trust_chain.pem

  ./rfc8572-agent.sh "$CLIENT_SERIAL_NUMBER" "$CLIENT_PASSWORD" "$CLIENT_CERT_KEY" "$CLIENT_CERT_EE" "$BOOTSVR_ADDRESS" "$BOOTSVR_PORT" "$BOOTSVR_TA_CERT"
  exit_code=$?
  echo "  ^-- rfc8572-agent exited with status code $exit_code"
}


main() {

  if [ ! -f pki/sztpd3/sbi/end-entity/my_cert.pem ]; then
    # note: the file tested for existance is arbitrary, but happens to be last cert signed...
    #echo 'File not found: pki/sztpd3/sbi/end-entity/my_cert.pem'  
    echo ''
    echo '    The PKI needs to be initialized first.'
    echo '        - `cd pki; make pki; cd -`'
    echo ''
    exit 1
  fi

  # create tempdir for files created during run 
  TEMPDIR=`mktemp -d`
  echo "Temporary directory for output files: $TEMPDIR"
  echo

  # strip out axes that cause `yanglint` to fail
  cp -R $YANG_DIR/ $TEMPDIR/new_yang-dir
  for i in `ls $TEMPDIR/new_yang-dir/wn-app\@*.yang $TEMPDIR/new_yang-dir/wn-sztpd-1\@*.yang`; do
    #echo "Stripping out any lines containing \"STRIP_4CLI\"..."
    # no need to restore, since in a copy dir...
    sed -i.bak '/STRIP_4CLI/d' $i
  done

  # update YANG_DIR var to point to new location
  YANG_DIR="$TEMPDIR/new_yang-dir"
  #echo "NEW YANG_DIR: $YANG_DIR"

  # set a couple global vars
  HEADER_FN="$TEMPDIR/header.dump"
  CURL="curl --silent --dump-header $HEADER_FN"

  echo "Creating instances..."
  create_sztpd_instance 1
  create_sztpd_instance 2
  create_sztpd_instance 3

  echo
  echo "Giving servers time to startup..."
  sleep 10

  echo
  echo "Configuring instances..."
  configure_sztpd_instance 1
  configure_sztpd_instance 2
  configure_sztpd_instance 3

  echo
  echo "Giving servers time to open their ports..."
  sleep 10

  echo
  echo "Testing instances..."
  test_sztpd_instance 1
  test_sztpd_instance 2
  test_sztpd_instance 3

  echo
  echo "Running simulator..."
  run_rfc8572_agent

  #echo
  #echo "Testing instances after..."
  #test_sztpd_instance_after 1
  #test_sztpd_instance_after 2
  #test_sztpd_instance_after 3

  echo
  echo 'Killing `sztpd` instances...'
  kill_sztpd_instance 1
  kill_sztpd_instance 2
  kill_sztpd_instance 3

  echo
  echo "Giving servers time to shutdown..."
  sleep 4

  echo
  echo 'Verifying instances are killed...'
  verify_instance_killed 1
  verify_instance_killed 2
  verify_instance_killed 3

  # remove tempdir
  rm -rf "$TEMPDIR"

  echo
  echo "All done!"
  exit 0
}


main "$@"

